@@ -14,27 +14,46 @@ SEXP Model__get_W(SEXP xp, int type = 0){
 }
 
 // [[Rcpp::export]]
-void Model__set_lower_bound(SEXP xp, SEXP bound_, int type = 0){
+void Model__set_direct_control(SEXP xp, bool direct = false, double direct_range_beta = 3.0, int max_iter = 100, double epsilon = 1e-4, bool select_one = true, bool trisect_once = false, 
+                        int max_eval = 0, bool mrdirect = false, int type = 0){
   glmmrType model(xp,static_cast<Type>(type));
-  std::vector<double> bound = as<std::vector<double> >(bound_);
   auto functor = overloaded {
     [](int) {}, 
-    [&bound](auto ptr){ptr->optim.set_bound(bound,true);}
+    [&](auto ptr){ptr->optim.set_direct_control(direct, direct_range_beta, max_iter, epsilon, select_one, trisect_once, max_eval, mrdirect);}
   };
   std::visit(functor,model.ptr);
 }
 
 // [[Rcpp::export]]
-void Model__set_upper_bound(SEXP xp, SEXP bound_, int type = 0){
+void Model__set_lbfgs_control(SEXP xp, double g_epsilon = 1e-8, int past = 3, double delta = 1e-8, int max_linesearch = 64, int type = 0){
   glmmrType model(xp,static_cast<Type>(type));
-  std::vector<double> bound = as<std::vector<double> >(bound_);
   auto functor = overloaded {
     [](int) {}, 
-    [&bound](auto ptr){ptr->optim.set_bound(bound,false);}
+    [&](auto ptr){ptr->optim.set_lbfgs_control(g_epsilon, past, delta, max_linesearch);}
   };
   std::visit(functor,model.ptr);
 }
 
+// [[Rcpp::export]]
+void Model__set_bound(SEXP xp, SEXP bound_, bool beta = true, bool lower = true, int type = 0){
+  glmmrType model(xp,static_cast<Type>(type));
+  std::vector<double> bound = as<std::vector<double> >(bound_);
+  if(beta){
+    auto functor = overloaded {
+      [](int) {}, 
+      [&](auto ptr){ptr->optim.set_bound(bound,lower);}
+    };
+    std::visit(functor,model.ptr);
+  } else {
+    auto functor = overloaded {
+      [](int) {}, 
+      [&](auto ptr){ptr->optim.set_theta_bound(bound,lower);}
+    };
+    std::visit(functor,model.ptr);
+  }
+  
+}
+
 // [[Rcpp::export]]
 void Model__print_instructions(SEXP xp, bool linpred, bool loglik, int type = 0){
   glmmrType model(xp,static_cast<Type>(type));
@@ -116,16 +135,6 @@ SEXP Model__log_likelihood(SEXP xp, int type = 0){
   return wrap(std::get<double>(S));
 }
 
-// [[Rcpp::export]]
-void Model__ml_theta(SEXP xp, int type = 0){
-  glmmrType model(xp,static_cast<Type>(type));
-  auto functor = overloaded {
-    [](int) {}, 
-    [](auto ptr){ptr->optim.ml_theta();}
-  };
-  std::visit(functor,model.ptr);
-}
-
 // [[Rcpp::export]]
 void Model__cov_set_nn(SEXP xp, int nn){
   XPtr<glmm_nngp> ptr(xp);
@@ -133,51 +142,191 @@ void Model__cov_set_nn(SEXP xp, int nn){
 }
 
 // [[Rcpp::export]]
-void Model__ml_beta(SEXP xp, int type = 0){
+void Model__test_lbfgs(SEXP xp, SEXP x){
+  XPtr<glmm> ptr(xp);
+  Eigen::VectorXd start = as<Eigen::VectorXd>(x);
+  Eigen::VectorXd grad(start.size());
+  grad.setZero();
+  double ll = ptr->optim.log_likelihood_beta_with_gradient(start,grad);
+  Rcpp::Rcout << "\nStart: " << start.transpose();
+  Rcpp::Rcout << "\nGradient: " << grad.transpose();
+  Rcpp::Rcout << "\nLog likelihood: " << ll;
+}
+
+// [[Rcpp::export]]
+void Model__test_lbfgs_theta(SEXP xp, SEXP x){
+  XPtr<glmm> ptr(xp);
+  Eigen::VectorXd start = as<Eigen::VectorXd>(x);
+  Eigen::VectorXd grad(start.size());
+  grad.setZero();
+  if(ptr->re.scaled_u_.cols() != ptr->re.u_.cols())ptr->re.scaled_u_.conservativeResize(NoChange,ptr->re.u_.cols());
+  ptr->re.scaled_u_ = ptr->model.covariance.Lu(ptr->re.u_);  
+  double ll = ptr->optim.log_likelihood_theta_with_gradient(start,grad);
+  Rcpp::Rcout << "\nStart: " << start.transpose();
+  Rcpp::Rcout << "\nGradient: " << grad.transpose();
+  Rcpp::Rcout << "\nLog likelihood: " << ll;
+}
+
+// [[Rcpp::export]]
+void Model__test_lbfgs_laplace(SEXP xp, SEXP x){
+  XPtr<glmm> ptr(xp);
+  Eigen::VectorXd start = as<Eigen::VectorXd>(x);
+  Eigen::VectorXd grad(start.size());
+  grad.setZero();
+  if(ptr->re.scaled_u_.cols() != ptr->re.u_.cols())ptr->re.scaled_u_.conservativeResize(NoChange,ptr->re.u_.cols());
+  ptr->re.scaled_u_ = ptr->model.covariance.Lu(ptr->re.u_);
+  double ll = ptr->optim.log_likelihood_laplace_beta_u_with_gradient(start,grad);
+  Rcpp::Rcout << "\nStart: " << start.transpose();
+  Rcpp::Rcout << "\nGradient: " << grad.transpose();
+  Rcpp::Rcout << "\nLog likelihood: " << ll;
+}
+
+// [[Rcpp::export]]
+void Model__ml_beta(SEXP xp, int algo = 0, int type = 0){
   glmmrType model(xp,static_cast<Type>(type));
   auto functor = overloaded {
     [](int) {}, 
-    [](auto ptr){ptr->optim.ml_beta();}
+    [&algo](auto ptr){
+      switch(algo){
+        case 1:
+          ptr->optim.template ml_beta<NEWUOA>();
+          break;
+        case 2:
+          ptr->optim.template ml_beta<LBFGS>();
+          break;
+        case 3:
+          ptr->optim.template ml_beta<DIRECT>();
+          break;
+        default:
+          ptr->optim.template ml_beta<BOBYQA>();
+          break;
+      }
+    }
   };
   std::visit(functor,model.ptr);
 }
 
 // [[Rcpp::export]]
-void Model__ml_all(SEXP xp, int type = 0){
+void Model__ml_theta(SEXP xp, int algo = 0, int type = 0){
   glmmrType model(xp,static_cast<Type>(type));
   auto functor = overloaded {
     [](int) {}, 
-    [](auto ptr){ptr->optim.ml_all();}
+    [&algo](auto ptr){
+      switch(algo){
+      case 1:
+        ptr->optim.template ml_theta<NEWUOA>();
+        break;
+      case 2:
+        ptr->optim.template ml_theta<LBFGS>();
+        break;
+      case 3:
+        ptr->optim.template ml_theta<DIRECT>();
+        break;
+      default:
+        ptr->optim.template ml_theta<BOBYQA>();
+      break;
+      }
+    }
   };
   std::visit(functor,model.ptr);
 }
 
 // [[Rcpp::export]]
-void Model__laplace_ml_beta_u(SEXP xp, int type = 0){
+void Model__ml_all(SEXP xp, int algo = 0, int type = 0){
   glmmrType model(xp,static_cast<Type>(type));
   auto functor = overloaded {
     [](int) {}, 
-    [](auto ptr){ptr->optim.laplace_ml_beta_u();}
+    [&algo](auto ptr){
+      switch(algo){
+      case 1:
+        ptr->optim.template ml_all<NEWUOA>();
+        break;
+      case 2:
+        Rcpp::stop("L-BGFS not available for full likelihood beta-theta joint optimisation.");
+        break;
+      case 3:
+        ptr->optim.template ml_all<DIRECT>();
+        break;
+      default:
+        ptr->optim.template ml_all<BOBYQA>();
+      break;
+      }
+    }
   };
   std::visit(functor,model.ptr);
 }
 
 // [[Rcpp::export]]
-void Model__laplace_ml_theta(SEXP xp, int type = 0){
+void Model__laplace_ml_beta_u(SEXP xp, int algo = 0, int type = 0){
   glmmrType model(xp,static_cast<Type>(type));
   auto functor = overloaded {
     [](int) {}, 
-    [](auto ptr){ptr->optim.laplace_ml_theta();}
+    [&algo](auto ptr){
+      switch(algo){
+      case 1:
+        ptr->optim.template laplace_ml_beta_u<NEWUOA>();
+        break;
+      case 2:
+        ptr->optim.template laplace_ml_beta_u<LBFGS>();
+        break;
+      case 3:
+        ptr->optim.template laplace_ml_beta_u<DIRECT>();
+        break;
+      default:
+        ptr->optim.template laplace_ml_beta_u<BOBYQA>();
+      break;
+      }
+    }
   };
   std::visit(functor,model.ptr);
 }
 
 // [[Rcpp::export]]
-void Model__laplace_ml_beta_theta(SEXP xp, int type = 0){
+void Model__laplace_ml_theta(SEXP xp, int algo = 0, int type = 0){
   glmmrType model(xp,static_cast<Type>(type));
   auto functor = overloaded {
     [](int) {}, 
-    [](auto ptr){ptr->optim.laplace_ml_beta_theta();}
+    [&algo](auto ptr){
+      switch(algo){
+      case 1:
+        ptr->optim.template laplace_ml_theta<NEWUOA>();
+        break;
+      case 2:
+        ptr->optim.template laplace_ml_theta<LBFGS>();
+        break;
+      case 3:
+        ptr->optim.template laplace_ml_theta<DIRECT>();
+        break;
+      default:
+        ptr->optim.template laplace_ml_theta<BOBYQA>();
+      break;
+      }
+    }
+  };
+  std::visit(functor,model.ptr);
+}
+
+// [[Rcpp::export]]
+void Model__laplace_ml_beta_theta(SEXP xp, int algo = 0, int type = 0){
+  glmmrType model(xp,static_cast<Type>(type));
+  auto functor = overloaded {
+    [](int) {}, 
+    [&algo](auto ptr){
+      switch(algo){
+      case 1:
+        ptr->optim.template laplace_ml_beta_theta<NEWUOA>();
+        break;
+      case 2:
+        Rcpp::stop("L-BGFS(-B) is not available for Laplace beta-theta optimisation");
+        break;
+      case 3:
+        ptr->optim.template laplace_ml_beta_theta<DIRECT>();
+        break;
+      default:
+        ptr->optim.template laplace_ml_beta_theta<BOBYQA>();
+      break;
+      }
+    }
   };
   std::visit(functor,model.ptr);
 }
@@ -489,17 +638,6 @@ SEXP Model__ZL(SEXP xp, int type = 0){
   return wrap(std::get<Eigen::MatrixXd>(S));
 }
 
-// [[Rcpp::export]]
-SEXP Model__hessian_numerical(SEXP xp, double tol = 1e-4, int type = 0){
-  glmmrType model(xp,static_cast<Type>(type));
-  auto functor = overloaded {
-    [](int) {  return returnType(0);}, 
-    [&tol](auto ptr){return returnType(ptr->optim.hessian_numerical(tol));}
-  };
-  auto S = std::visit(functor,model.ptr);
-  return wrap(std::get<Eigen::MatrixXd>(S));
-}
-
 // [[Rcpp::export]]
 SEXP Model__xb(SEXP xp, int type = 0){
   glmmrType model(xp,static_cast<Type>(type));
